<?php

/**
 * EEE_Infusionsoft_Attendee
 *
 * @package             Event Espresso
 * @subpackage
 * @author              Mike Nelson
 */
class EEE_Infusionsoft_Attendee extends EEE_Base_Class
{
    const extra_meta_IS_contact_ID = 'IS_contact_ID';

    const extra_meta_IS_group_IDs = 'IS_group_IDs';


    /**
     * Array of attendee IDs as keys with their related IS contact ID as the value.
     *
     * @var array
     */
    private $IS_contact_cache = [];


    public function __construct()
    {
        $this->_model_name_extended = 'Attendee';
        parent::__construct();
    }


    public function ext_sync_to_infusionsoft()
    {
        try {
            // only sync attendees actually in the DB
            if (! $this->_->ID() || EED_Infusionsoft::synced_on_this_request($this->_)) {
                return false;
            }
            $IS_contact_ID  = $this->_->get_IS_contact_ID();
            $isdk           = EED_Infusionsoft::infusionsoft_connection();
            $StreetAddress1 = $this->_->address();
            $StreetAddress2 = $this->_->address2();
            $City           = $this->_->city();
            $State          = $this->_->state_name();
            $PostalCode     = $this->_->zip();
            $Phone1         = $this->_->phone();
            $Country        = $this->_->country_name();


            // map the EE4 Attendee fields onto IS contact data
            $contact_data = [
                'FirstName'      => $this->_->fname(),
                'LastName'       => $this->_->lname(),
                'Email'          => $this->_->email(),
                'StreetAddress1' => $StreetAddress1,
                'StreetAddress2' => $StreetAddress2,
                'City'           => $City,
                'State'          => $State,
                'PostalCode'     => $PostalCode,
                'Phone1'         => $Phone1,
                'Country'        => $Country,
            ];

            // also record their answers to custom questions
            $mapped_custom_questions = EEM_Question::instance()->get_all(
                [
                    [
                        'Extra_Meta.EXM_key'    => EEE_Infusionsoft_Question::extra_meta_IS_field_name,
                        'Extra_Meta.EXM_value'  => ['!=', ''],
                        'Extra_Meta.EXM_value*' => ['IS_NOT_NULL'],
                    ],
                ]
            );
            foreach ($mapped_custom_questions as $question) {
                $answer_to_that_question = EEM_Answer::instance()->get_var(
                    [
                        [
                            'QST_ID' => $question->ID(),
                        ],
                        'order_by' => [
                            'REG_ID' => 'DESC',
                        ],
                    ],
                    'ANS_value'
                );
                if ($question->type() === EEM_Question::QST_type_date) {
                    // @todo: convert it form a format like "November 10, 2017" to EED_Infusionsoft::IS_datetime_format
                    $datetime = DateTime::createFromFormat(get_option('date_format'), $answer_to_that_question);
                    if (! $datetime instanceof DateTime) {
                        // if we didn't get a datetime, it's an invalid format.
                        // it's possible it was a valid format at one point, so just leave it alone
                        continue;
                    } else {
                        $answer_to_that_question = $datetime->format(EED_Infusionsoft::IS_datetime_format);
                    }
                }
                // use dynamically added method in EEE_Infusionsoft_Question
                $contact_data[ $question->get_IS_field_db_name() ] = $answer_to_that_question;
            }

            if (empty($StreetAddress1)) {
                unset($contact_data['StreetAddress1']);
            }
            if (empty($StreetAddress2)) {
                unset($contact_data['StreetAddress2']);
            }
            if (empty($City)) {
                unset($contact_data['City']);
            }
            $state_id = $this->_->state_ID();
            if (empty($state_id)) {
                unset($contact_data['State']);
            }
            if (empty($PostalCode)) {
                unset($contact_data['PostalCode']);
            }
            if (empty($Phone1)) {
                unset($contact_data['Phone1']);
            }
            $country_id = $this->_->country_ID();
            if (empty($country_id)) {
                unset($contact_data['Country']);
            }

            /**
             * Filters the fields sync'd from an EE Attendee to its corresponding IS contact
             *
             * @param array       $attendee_data keys are Contact field names mentioned on https://developer.infusionsoft.com/docs/read/Table_Documentation
             * @param EE_Attendee $attendee      the EE4 attendee being sync'd to IS
             */
            $contact_data = apply_filters(
                'FHEE__EED_Infusionsoft__save_infusionsoft_attendee__extra_attendee_data',
                $contact_data,
                $this->_
            );

            if ($IS_contact_ID) {
                try {
                    $previous_contact_data = $isdk->data()->load("Contact", $IS_contact_ID, ['ContactNotes']);
                } catch (Exception $e) {
                    EE_PMT_Infusionsoft_Onsite::errorLog(
                        sprintf(
                            __(
                                'Infusionsoft error. Could not load contact with ID %1$s from EE attendee %2$s. Error was %3$s',
                                'event_espresso'
                            ),
                            $IS_contact_ID,
                            $this->_->ID(),
                            $e->getMessage()
                        ),
                        [],
                        EEM_Transaction::instance()->get_one_by_ID($this->_->ID()) ?? null
                    );
                    return false;
                }
                $old_contact_notes = $previous_contact_data['ContactNotes'] ?? '';
                $new_contact_data  = $old_contact_notes . '\r\n' . sprintf(
                    __(
                        'Account was updated on %s via your event registration system (Event Espresso %s).',
                        'event_espresso'
                    ),
                    date("Y-m-d", time()),
                    espresso_version(),
                    $this->_->ID()
                );
                $contact_data['ContactNotes'] = $new_contact_data;
                try {
                    $success = $isdk->data()->update('Contact', $IS_contact_ID, $contact_data);
                } catch (Exception $e) {
                    EE_PMT_Infusionsoft_Onsite::errorLog(
                        sprintf(
                            __(
                                'Did not successfully updated IS contact \'%1$s\' with data %2$s. Infusionsoft API response was %3$s',
                                'event_espresso'
                            ),
                            $IS_contact_ID,
                            json_encode($contact_data),
                            $e->getMessage()
                        ),
                        [],
                        EEM_Transaction::instance()->get_one_by_ID($this->_->ID()) ?? null
                    );
                }
                $newly_synced = false;
            } else {
                // Create a variable for the contact id
                $IS_contact_ID = EED_Infusionsoft::addWithDupCheck($contact_data);
                $this->_->set_IS_contact_ID($IS_contact_ID);
                if ($IS_contact_ID) {
                    try {
                        $opt_in_success = $isdk->emails('xml')->optIn($contact_data['Email'], '');
                    } catch (Exception $e) {
                        EE_PMT_Infusionsoft_Onsite::errorLog(
                            sprintf(
                                __(
                                    'Opt in failed for attendee with ID %1$s. Response from IS was %2$s',
                                    'event_espresso'
                                ),
                                $this->_->ID(),
                                $e->getMessage()
                            ),
                            [],
                            EEM_Transaction::instance()->get_one_by_ID($this->_->ID()) ?? null
                        );
                    }
                    $new_contact_notes = sprintf(
                        __(
                            'Account was created on %s via your event registration system (Event Espresso %s). Attendee ID: %d ',
                            'event_espresso'
                        ),
                        date("Y-m-d", time()),
                        espresso_version(),
                        $this->_->ID()
                    );
                    try {
                        $success = $isdk->data()->update('Contact', $IS_contact_ID, ['ContactNotes' => $new_contact_notes]);
                    } catch (Exception $e) {
                        EE_PMT_Infusionsoft_Onsite::errorLog(
                            sprintf(
                                __(
                                    'Did not successfully add IS contact to IS from attendee with ID %1$s. Response from IS was %2$s',
                                    'event_espresso'
                                ),
                                $this->_->ID(),
                                $e->getMessage()
                            ),
                            [],
                            EEM_Transaction::instance()->get_one_by_ID($this->_->ID()) ?? null
                        );
                    }
                }
                $newly_synced = true;
            }
            /**
             * Action performed after syncing an EE Attendee to an IS contact successfully
             *
             * @param EE_Attendee $attendee
             * @param string      $IS_contact_ID the ID of the IS contact in IS
             * @param array       $contact_data  sent to IS
             * @param boolean     $newly_synced  whether we just sync'd for the first time or not
             */
            do_action(
                'AHEE__EEE_Infusionsoft_Attendee__sync_to_infusionsoft__end',
                $this->_,
                $IS_contact_ID,
                $contact_data,
                $newly_synced
            );
        } catch (Exception $e) {
            EE_PMT_Infusionsoft_Onsite::errorLog(
                sprintf(
                    __(
                        'Exception thrown! Could not sync Event Espresso Attendee (%1$s) to Infusionsoft Contact because: %2$s',
                        'event_espresso'
                    ),
                    $this->_->ID(),
                    $e->getMessage() . $e->getTraceAsString()
                ),
                [],
                EEM_Transaction::instance()->get_one_by_ID($this->_->ID()) ?? null
            );
            return false;
        }
        EED_Infusionsoft::mark_model_obj_synced($this->_);
        return true;
    }


    /**
     * Pulls the IS_contact_ID assigned to the contact and checks if its valid
     * If not this tries to pull a matching contact from IS using the primary attendee details
     *
     * @return int|null
     */
    public function ext_get_IS_contact_ID()
    {
        // Check if we already have a cached IS_contact_ID.
        if (array_key_exists($this->_->ID(), $this->IS_contact_cache)) {
            return $this->IS_contact_cache[ $this->_->ID() ];
        }

        // No cached value so check if there is an IS_contact_ID set on the attendee.
        $IS_contact_ID = (int) $this->_->get_extra_meta(self::extra_meta_IS_contact_ID, true);
        if (empty($IS_contact_ID)) {
            // No previously saved IS contact ID
            return null;
        }

        // So we have an IS_contact_ID, check it's a valid contact in Infusionsoft.
        $isdk = EED_Infusionsoft::infusionsoft_connection();
        try {
            $is_valid_IS_ID = $isdk->data()->load("Contact", $IS_contact_ID, ['Id']);
        } catch (Exception $e) {
            $is_valid_IS_ID = [];
        }
        if (EED_Infusionsoft::is_IS_error($is_valid_IS_ID)) {
            // IS_contact_ID is not a valid contact in IS so attempt to pull a contact that matches the primary attendee
            $contact_data     = [
                'FirstName' => $this->_->fname(),
                'LastName'  => $this->_->lname(),
                'Email'     => $this->_->email(),
            ];
            try {
                $matching_contact = $isdk->data()->query('Contact', 1, 0, $contact_data, ['Id'], '', true);
            } catch (Exception $e) {
                // No matching IS contact can be found.
                EE_PMT_Infusionsoft_Onsite::errorLog(
                    sprintf(
                        __(
                            'The original IS contact id (%1$s) liked to Attendee %2$s is not valid and no matching contact can be found, a new contact will be created.',
                            'event_espresso'
                        ),
                        $IS_contact_ID,
                        $this->_->ID()
                    ),
                    [],
                    EEM_Transaction::instance()->get_one_by_ID($this->_->ID()) ?? null
                );
                return null;
            }
            // We have a matching contact from IS, use that contact ID.
            $first_contact = array_shift($matching_contact);
            if (! isset($first_contact['Id'])) {
                return null;
            }
            $IS_contact_ID = (int) $first_contact['Id'];
            // Update the IS_contact_ID value set on the attendee.
            $this->_->set_IS_contact_ID($IS_contact_ID);
        }
        // If we made it here $IS_contact_ID is now valid, cache the value.
        $this->IS_contact_cache[ $this->_->ID() ] = $IS_contact_ID;
        return $IS_contact_ID;
    }


    public function ext_set_IS_contact_ID($new_id)
    {
        $this->_->update_extra_meta(self::extra_meta_IS_contact_ID, $new_id);
    }


    /**
     * Gets the locally saved array of strings whose values are group IDs
     * we know we've assigned this registrant's contact to
     *
     * @return string[] numerically indexed with values of IS group IDs
     */
    public function ext_groups_synced()
    {
        return $this->_->get_extra_meta(self::extra_meta_IS_group_IDs, false, []);
    }


    public function ext_save_group_synced($group_id)
    {
        return $this->_->add_extra_meta(self::extra_meta_IS_group_IDs, $group_id, false);
    }


    /**
     * Returns an array of all the IS groups in $group_ids that we have not yet
     * sync'ed this contact to
     *
     * @param string[] $group_ids
     * @return string[] all the IS groups we should add this contact to in IS
     */
    public function ext_determine_groups_to_sync($group_ids)
    {
        return array_diff($group_ids, $this->_->groups_synced());
    }
}
