<?php

use EventEspresso\core\exceptions\ExceptionLogger;

/**
 * Class  EE_PMT_Infusionsoft_Onsite
 *
 * @package         Event Espresso
 * @subpackage      eea-infusionsoft-payments
 * @author          Event Espresso
 */
class EE_PMT_Infusionsoft_Onsite extends EE_PMT_Base
{
    /**
     * @param EE_Payment_Method|null $pm_instance
     * @throws EE_Error
     * @throws ReflectionException
     */
    public function __construct($pm_instance = null)
    {
        require_once(EE_INFUSIONSOFT_PATH . 'payment_methods/Infusionsoft_Onsite/EEG_Infusionsoft_Onsite.gateway.php');
        $this->_gateway        = new EEG_Infusionsoft_Onsite();
        $this->_pretty_name    = __('Infusionsoft', 'event_espresso');
        $this->_template_path  = EE_INFUSIONSOFT_PATH . 'payment_methods/Infusionsoft_Onsite/templates/';
        $this->_requires_https = true;
        parent::__construct($pm_instance);
    }


    /**
     * Adds the help tab.
     *
     * @return array
     * @see EE_PMT_Base::help_tabs_config()
     */
    public function help_tabs_config()
    {
        return [
            $this->get_help_tab_name() => [
                'title'         => esc_html__('Infusionsoft Payments Settings', 'event_espresso'),
                'filename'      => 'payment_methods_infusionsoft',
                'template_args' => [
                    'site_url' => get_admin_url(
                        get_current_blog_id(),
                        'admin.php?page=espresso_infusionsoft'
                    ),
                ],
            ],
        ];
    }


    /**
     * Gets the form for all the settings related to this payment method.
     *
     * @return EE_Payment_Method_Form
     */
    public function generate_new_settings_form()
    {
        return new EE_Infusionsoft_PM_Form($this);
    }


    /**
     *  Creates a billing form for this payment method type.
     *
     * @param EE_Transaction|null $transaction
     * @return EE_Billing_Info_Form|null
     * @throws EE_Error
     * @throws ReflectionException
     */
    public function generate_new_billing_form(EE_Transaction $transaction = null)
    {
        $allowed_types = $this->_pm_instance->get_extra_meta('credit_card_types', true);
        // If allowed types is a string or empty array or null.
        if (empty($allowed_types)) {
            $allowed_types = [];
        }

        return new EE_Billing_Attendee_Info_Form(
            $this->_pm_instance,
            [
                'name'        => 'Infusionsoft_Payment_Info_Form',
                'subsections' => [
                    'card_type'    => new EE_Select_Input(
                        array_intersect_key(
                            EE_PMT_Infusionsoft_Onsite::card_types_supported(),
                            array_flip($allowed_types)
                        ),
                        [
                            'html_class'      => 'eea-infusionsoft-billing-form-card-cvv',
                            'html_label_text' => __('Card Type', 'event_espresso'),
                            'required'        => true,
                        ]
                    ),
                    'name_on_card' => new EE_Text_Input(
                        [
                            'html_class'      => 'eea-infusionsoft-billing-form-name-on-card',
                            'html_label_text' => __('Name On Card', 'event_espresso'),
                            'required'        => true,
                        ]
                    ),
                    'credit_card'  => new EE_Credit_Card_Input(
                        [
                            'html_class'      => 'eea-infusionsoft-billing-form-credit-card',
                            'html_label_text' => __('Card Number', 'event_espresso'),
                            'required'        => true,
                        ]
                    ),
                    'exp_month'    => new EE_Credit_Card_Month_Input(
                        true,
                        [
                            'html_class'      => 'eea-infusionsoft-billing-form-exp-month',
                            'html_label_text' => __('Expiry Month', 'event_espresso'),
                            'required'        => true,
                        ]
                    ),
                    'exp_year'     => new EE_Credit_Card_Year_Input(
                        [
                            'html_class'      => 'eea-infusionsoft-billing-form-exp-year',
                            'html_label_text' => __('Expiry Year', 'event_espresso'),
                            'required'        => true,
                        ]
                    ),
                    'card_cvv'     => new EE_CVV_Input(
                        [
                            'html_class'      => 'eea-infusionsoft-billing-form-card-cvv',
                            'html_label_text' => __('CVV', 'event_espresso'),
                            'required'        => true,
                        ]
                    ),
                ],
            ]
        );
    }


    /**
     * Allow other code to filter the biling info sent to Infusionsoft, because
     * different gateways setup with IS may have different requirements
     *
     * @param EE_Billing_Info_Form|null $billing_form
     * @return array
     * @throws EE_Error
     */
    protected function _get_billing_values_from_form($billing_form)
    {
        return apply_filters(
            'FHEE__EE_PMT_Infusionsoft_Onsite___get_billing_values_from_form',
            parent::_get_billing_values_from_form($billing_form),
            $billing_form,
            $this
        );
    }


    /**
     * Returns an array of all the payment cards possibly supported by Infusionsoft Payments.
     * Keys are their values, values are their pretty names.
     *
     * @return array
     */
    public static function card_types_supported(): array
    {
        return [
            'Visa'       => __('Visa', 'event_espresso'),
            'MasterCard' => __('MasterCard', 'event_espresso'),
            'Amex'       => __('American Express', 'event_espresso'),
            'Discover'   => __('Discover', 'event_espresso'),
        ];
    }


    /**
     * Log a PM error.
     *
     * @param string                 $error_message
     * @param array                  $data
     * @param mixed                  $object_logged
     * @param EE_Payment_Method|null $keap_pm
     * @return bool
     */
    public static function errorLog(
        string $error_message = '',
        array $data = [],
        $object_logged = null,
        ?EE_Payment_Method $keap_pm = null
    ): bool {
        $default_msg = $error_message ?? 'Keap integration error';
        $data        = $data ? EE_PMT_Infusionsoft_Onsite::arrayCleanData($data) : $data;
        try {
            if (! $keap_pm instanceof EE_Payment_Method) {
                // Default to a default payment method slug.
                $keap_pm = EEM_Payment_Method::instance()->get_one_by_slug('infusionsoft_onsite');
            }
            if (! $keap_pm instanceof EE_Payment_Method) {
                error_log($error_message);
                return false;
            }
            $keap_gateway = $keap_pm->type_obj()->get_gateway();
            if ($keap_gateway instanceof EE_Gateway) {
                $keap_gateway->log([$default_msg => $data], $object_logged);
            }
        } catch (ReflectionException | EE_Error $error) {
            new ExceptionLogger($error);
            return false;
        }
        return true;
    }


    /**
     * Clear out the provided sensitive data.
     *
     * @param array $data Data to be logged
     * @return array
     */
    public static function arrayCleanData($data)
    {
        // Data keys to be removed.
        $sensitive_data = array(
            'credit_card',
            'exp_month',
            'exp_year',
            'card_cvv',
            'CardNumber',
            'ExpirationMonth',
            'ExpirationYear',
            'CVV2',
            'NameOnCard'
        );
        foreach ($data as $key => $value) {
            if (in_array($key, $sensitive_data)) {
                unset($data[ $key ]);
            } elseif (is_array($value) && ! empty($value)) {
                $data[ $key ] = EE_PMT_Infusionsoft_Onsite::arrayCleanData($value);
            }
        }
        return $data;
    }
}
